package adapterclient

import (
	"errors"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

const SuccessExitCode = 0

type Adapter struct {
	ExternalBinPath string
	CommandRunner   CommandRunner
}

func SanitiseForJSON(properties serviceadapter.Properties) serviceadapter.Properties {
	propertiesToReturn := serviceadapter.Properties{}

	for k, v := range properties {
		propertiesToReturn[k] = sanitiseValueForJSON(v)
	}

	return propertiesToReturn
}

// structures parsed from YAML contain map[interface{}]interface{}, but JSON marshalling requires map[string]interface{}, so we have to walk the structure and convert any instances of the former to the latter
func sanitiseValueForJSON(obj interface{}) interface{} {
	switch castObj := obj.(type) {

	case map[interface{}]interface{}:
		sanitisedMap := map[string]interface{}{}
		for key, value := range castObj {
			sanitisedMap[key.(string)] = sanitiseValueForJSON(value)
		}
		return sanitisedMap

	case []interface{}:
		var sanitisedSlice []interface{}
		for _, value := range castObj {
			sanitisedSlice = append(sanitisedSlice, sanitiseValueForJSON(value))
		}
		return sanitisedSlice

	default:
		return obj

	}
}

type errorFactory struct {
	externalBinPath string
	stdout, stderr  string
}

func newErrorFactory(externalBinPath string, stdout, stderr []byte) errorFactory {
	return errorFactory{
		externalBinPath: externalBinPath,
		stdout:          string(stdout),
		stderr:          string(stderr),
	}
}

func (e errorFactory) adapterError(err error) error {
	return fmt.Errorf("an error occurred running external service adapter at %s: '%s'. Stdout: '%s', stderr: '%s'", e.externalBinPath, err, e.stdout, e.stderr)
}
func (e errorFactory) incorrectDeploymentNameError(expectedName, actualName string) error {
	return fmt.Errorf("external service adapter generated manifest with an incorrect deployment name at %s. expected name: '%s', returned name: '%s', stderr: '%s'", e.externalBinPath, expectedName, actualName, e.stderr)
}
func (e errorFactory) invalidYamlError() error {
	return fmt.Errorf("external service adapter generated manifest that is not valid YAML at %s. stderr: '%s'", e.externalBinPath, e.stderr)
}
func (e errorFactory) invalidVersionError(version string) error {
	return fmt.Errorf("external service adapter generated manifest with an incorrect version at %s. expected exact version but returned version: '%s', stderr: '%s'", e.externalBinPath, version, e.stderr)
}
func (e errorFactory) exitCodeError(err error, exitCode int) error {
	messageForOperator := fmt.Sprintf("external service adapter exited with %d at %s: Stdout: '%s', stderr: '%s', exitError: '%s'", exitCode, e.externalBinPath, e.stdout, e.stderr, err)
	if e.stdout == "" {
		return errors.New(messageForOperator)
	}
	return broker.NewDisplayableError(errors.New(e.stdout), errors.New(messageForOperator))
}
