package adapterclient_test

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient/fake_command_runner"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("dashboard url", func() {
	const externalBinPath = "/thing"

	var (
		a          *adapterclient.Adapter
		cmdRunner  *fake_command_runner.FakeCommandRunner
		logs       *bytes.Buffer
		logger     *log.Logger
		instanceID string
		plan       serviceadapter.Plan
		manifest   []byte

		actualDashboardUrl string
		actualError        error
	)

	BeforeEach(func() {
		logs = new(bytes.Buffer)
		logger = log.New(io.MultiWriter(GinkgoWriter, logs), "[unit-tests] ", log.LstdFlags)
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &adapterclient.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
		}
		plan = serviceadapter.Plan{
			Properties: serviceadapter.Properties{
				"foo": "bar",
				"baz": map[string]interface{}{
					"qux": "quux",
				},
			},
		}

		manifest = []byte("a manifest")

		cmdRunner.RunReturns([]byte(""), []byte(""), intPtr(adapterclient.SuccessExitCode), nil)

		instanceID = "the-instance-id"
		manifest = []byte("a-manifest")

		dashboardUrlJSON := []byte(`{ "dashboard_url": "https://someurl.com"}`)

		cmdRunner.RunReturns(dashboardUrlJSON, []byte("I'm stderr"), intPtr(adapterclient.SuccessExitCode), nil)
	})

	JustBeforeEach(func() {
		actualDashboardUrl, actualError = a.GenerateDashboardUrl(instanceID, plan, manifest, logger)
	})

	It("invokes external dashboard url generator with serialised params", func() {
		Expect(cmdRunner.RunCallCount()).To(Equal(1))
		planJson, err := json.Marshal(plan)
		Expect(err).NotTo(HaveOccurred())
		argsPassed := cmdRunner.RunArgsForCall(0)
		Expect(argsPassed).To(ConsistOf(externalBinPath, "dashboard-url", instanceID, string(planJson), string(manifest)))
	})

	It("returns a dashboard url", func() {
		Expect(actualDashboardUrl).To(Equal("https://someurl.com"))
	})

	It("has no error", func() {
		Expect(actualError).NotTo(HaveOccurred())
	})

	It("logs adapter stderr", func() {
		Expect(logs).To(ContainSubstring("I'm stderr"))
	})

	Context("when plan properties are formatted as map[interface][interface]", func() {
		BeforeEach(func() {
			plan = serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
		})

		It("converts plan properties to be json serializable", func() {
			Expect(actualError).NotTo(HaveOccurred())
			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0)

			convertedPlan := serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[string]interface{}{
						"qux": "quux",
					},
				},
			}

			convertedPlanJson, err := json.Marshal(convertedPlan)
			Expect(err).NotTo(HaveOccurred())
			Expect(argsPassed).To(ConsistOf(externalBinPath, "dashboard-url", instanceID, string(convertedPlanJson), string(manifest)))
		})
	})

	Context("adapter did not implement the dashboard url subcommand", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(""), []byte("not impemented"), intPtr(serviceadapter.NotImplementedExitCode), nil)
		})

		It("returns the subcommand not implemented error", func() {
			Expect(actualError).To(HaveOccurred())
			Expect(actualError).To(BeAssignableToTypeOf(broker.NewServiceAdapterCommandNotImplementedError(fmt.Errorf("no relvent"))))
		})
	})

	Context("when the external service adapter fails, without an exit code", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
		})

		It("returns an error", func() {
			Expect(actualError).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
		})
	})

	Context("when the external service adapter fails", func() {
		Context("when there is a operator error message and a user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an DisplayableError", func() {
				commandError, ok := actualError.(broker.DisplayableError)
				Expect(ok).To(BeTrue(), "error should be an DisplayableError")
				Expect(actualError.Error()).To(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
				Expect(commandError.ErrorForCFUser()).To(MatchError("I'm stdout"))
			})
		})

		Context("when there is a operator error message and no user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte{}, []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an error", func() {
				_, ok := actualError.(broker.DisplayableError)
				Expect(ok).To(BeFalse(), "error should not be an DisplayableError")
				Expect(actualError).To(MatchError(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: '', stderr: 'I'm stderr'")))
			})
		})
	})

	Context("cannot deserialise response", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("invalid json"), []byte("I'm stderr"), intPtr(adapterclient.SuccessExitCode), nil)
		})

		It("returns an error", func() {
			Expect(actualError).To(HaveOccurred())
		})
	})
})
