package adapterclient_test

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient/fake_command_runner"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("external service adapter", func() {
	const externalBinPath = "/thing"

	var (
		a                  *adapterclient.Adapter
		cmdRunner          *fake_command_runner.FakeCommandRunner
		logs               *bytes.Buffer
		logger             *log.Logger
		bindingID          string
		deploymentTopology bosh.BoshVMs
		manifest           []byte
		requestParams      map[string]interface{}

		adapterBinding   serviceadapter.Binding
		createBindingErr error
	)

	BeforeEach(func() {
		logs = new(bytes.Buffer)
		logger = log.New(io.MultiWriter(GinkgoWriter, logs), "[unit-tests] ", log.LstdFlags)
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &adapterclient.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
		}
		cmdRunner.RunReturns([]byte(`{
  				"credentials": {
  					"username": "user1",
  					"password": "reallysecret"
  				},
  				"syslog_drain_url": "some-url",
  				"route_service_url": "route-url"
  			}`), []byte(""), intPtr(adapterclient.SuccessExitCode), nil)

		bindingID = "the-binding"
		deploymentTopology = bosh.BoshVMs{"the-deployment": []string{"a-vm"}}
		manifest = []byte("a-manifest")
		requestParams = map[string]interface{}{"foo": "bar"}
	})

	JustBeforeEach(func() {
		adapterBinding, createBindingErr = a.CreateBinding(bindingID, deploymentTopology, manifest, requestParams, logger)
	})

	It("invokes external binding creator with serialised params", func() {
		serialisedVMs, err := json.Marshal(deploymentTopology)
		Expect(err).NotTo(HaveOccurred())

		serialisedRequestParams, err := json.Marshal(requestParams)
		Expect(err).NotTo(HaveOccurred())

		Expect(cmdRunner.RunCallCount()).To(Equal(1))
		argsPassed := cmdRunner.RunArgsForCall(0)
		Expect(argsPassed).To(ConsistOf(externalBinPath, "create-binding", bindingID, string(serialisedVMs), string(manifest), string(serialisedRequestParams)))
	})

	Context("when the external adapter succeeds", func() {
		It("returns no error", func() {
			Expect(createBindingErr).ToNot(HaveOccurred())
		})

		It("returns the service-specific binding output", func() {
			Expect(adapterBinding).To(Equal(
				serviceadapter.Binding{
					Credentials:     map[string]interface{}{"username": "user1", "password": "reallysecret"},
					SyslogDrainURL:  "some-url",
					RouteServiceURL: "route-url",
				},
			))
		})
	})

	Context("when the external service adapter fails", func() {
		Context("when there is a operator error message and a user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an DisplayableError", func() {
				commandError, ok := createBindingErr.(broker.DisplayableError)
				Expect(ok).To(BeTrue(), "error should be an DisplayableError")
				Expect(createBindingErr.Error()).To(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
				Expect(commandError.ErrorForCFUser()).To(MatchError("I'm stdout"))
			})
		})

		Context("when there is a operator error message and no user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte{}, []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an error", func() {
				_, ok := createBindingErr.(broker.DisplayableError)
				Expect(ok).To(BeFalse(), "error should not be an DisplayableError")
				Expect(createBindingErr).To(MatchError(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: '', stderr: 'I'm stderr'")))
			})
		})
	})

	Context("when the external adapter fails with exit code 49", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.BindingAlreadyExistsErrorExitCode), nil)
		})

		It("returns an error", func() {
			Expect(createBindingErr).To(BeAssignableToTypeOf(serviceadapter.BindingAlreadyExistsError{}))
			Expect(createBindingErr).To(MatchError(ContainSubstring("external service adapter exited with 49 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
		})
	})

	Context("when the external adapter fails with exit code 42", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.AppGuidNotProvidedErrorExitCode), nil)
		})

		It("returns an error", func() {
			Expect(createBindingErr).To(BeAssignableToTypeOf(serviceadapter.AppGuidNotProvidedError{}))
			Expect(createBindingErr).To(MatchError(ContainSubstring("external service adapter exited with 42 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
		})
	})

	Context("when the external adapter fails with exit code 10", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.NotImplementedExitCode), nil)
		})

		It("returns an error", func() {
			Expect(createBindingErr.Error()).To(ContainSubstring("service adapter has not implemented create-binding command: error: external service adapter exited with 10 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
		})
	})

	Context("when the external adapter fails with no exit code", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
		})

		It("returns an error", func() {
			Expect(createBindingErr).To(Equal(fmt.Errorf("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
		})
	})

})
