package adapterclient_test

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient"
	"github.com/pivotal-cf/on-demand-service-broker/adapterclient/fake_command_runner"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

var _ = Describe("external service adapter", func() {
	const externalBinPath = "/thing"

	var (
		a                  *adapterclient.Adapter
		cmdRunner          *fake_command_runner.FakeCommandRunner
		logs               *bytes.Buffer
		logger             *log.Logger
		bindingID          string
		deploymentTopology bosh.BoshVMs
		manifest           []byte
		requestParams      map[string]interface{}

		deleteBindingError error
	)

	BeforeEach(func() {
		logs = new(bytes.Buffer)
		logger = log.New(io.MultiWriter(GinkgoWriter, logs), "[unit-tests] ", log.LstdFlags)
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &adapterclient.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
		}
		cmdRunner.RunReturns([]byte(""), []byte(""), intPtr(adapterclient.SuccessExitCode), nil)

		bindingID = "the-binding"
		deploymentTopology = bosh.BoshVMs{}
		manifest = []byte("a-manifest")
		requestParams = map[string]interface{}{
			"plan_id":    "some-plan-id",
			"service_id": "some-service-id",
		}
	})

	JustBeforeEach(func() {
		deleteBindingError = a.DeleteBinding(bindingID, deploymentTopology, manifest, requestParams, logger)
	})

	It("invokes external executable with params to delete binding", func() {
		serialisedBoshVMs, err := json.Marshal(deploymentTopology)
		Expect(err).NotTo(HaveOccurred())

		serialisedRequestParams, err := json.Marshal(requestParams)
		Expect(err).NotTo(HaveOccurred())

		Expect(cmdRunner.RunCallCount()).To(Equal(1))
		argsPassed := cmdRunner.RunArgsForCall(0)
		Expect(argsPassed).To(ConsistOf(externalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest), string(serialisedRequestParams)))
	})

	Context("when the external adapter succeeds", func() {
		It("returns no error", func() {
			Expect(deleteBindingError).ToNot(HaveOccurred())
		})
	})

	Context("when the external adapter fails with no exit code", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
		})
		It("returns an error", func() {
			Expect(deleteBindingError).To(Equal(fmt.Errorf("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
		})
	})

	Context("when the external adapter fails", func() {
		Context("when there is a operator error message and a user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an DisplayableError", func() {
				commandError, ok := deleteBindingError.(broker.DisplayableError)
				Expect(ok).To(BeTrue(), "error should be an DisplayableError")
				Expect(commandError.Error()).To(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
				Expect(commandError.ErrorForCFUser()).To(MatchError("I'm stdout"))
			})
		})

		Context("when there is a operator error message and no user error message", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte{}, []byte("I'm stderr"), intPtr(serviceadapter.ErrorExitCode), nil)
			})

			It("returns an error", func() {
				Expect(deleteBindingError).NotTo(BeAssignableToTypeOf(broker.DisplayableError{}))
				Expect(deleteBindingError).To(MatchError(ContainSubstring("external service adapter exited with 1 at /thing: Stdout: '', stderr: 'I'm stderr'")))
			})
		})
	})

	Context("when the external adapter fails with exit code 10", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.NotImplementedExitCode), nil)
		})

		It("returns an error", func() {
			Expect(deleteBindingError.Error()).To(ContainSubstring("external service adapter exited with 10 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			Expect(deleteBindingError).To(MatchError(ContainSubstring("service adapter has not implemented delete-binding command")))
		})
	})

	Context("when the external adapter fails with exit code 41", func() {
		BeforeEach(func() {
			cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(serviceadapter.BindingNotFoundErrorExitCode), nil)
		})

		It("returns an error", func() {
			Expect(deleteBindingError).To(BeAssignableToTypeOf(serviceadapter.BindingNotFoundError{}))
			Expect(deleteBindingError).To(MatchError(ContainSubstring("external service adapter exited with 41 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
		})
	})
})
