package adapterclient

import (
	"encoding/json"
	"fmt"
	"log"
	"strings"

	yaml "gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

type manifest struct {
	Name     string
	Releases []struct {
		Version string
	}
	Stemcells []struct {
		Version string
	}
}

type manifestValidator struct {
	deploymentName string
}

func (a *Adapter) GenerateManifest(serviceDeployment serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan, logger *log.Logger) ([]byte, error) {
	errs := func(err error) ([]byte, error) {
		return nil, err
	}

	serialisedServiceDeployment, err := json.Marshal(serviceDeployment)
	if err != nil {
		return errs(err)
	}

	plan.Properties = SanitiseForJSON(plan.Properties)
	serialisedPlan, err := json.Marshal(plan)
	if err != nil {
		return errs(err)
	}

	serialisedRequestParams, err := json.Marshal(requestParams)
	if err != nil {
		return errs(err)
	}
	if previousPlan != nil {
		previousPlan.Properties = SanitiseForJSON(previousPlan.Properties)
	}

	serialisedPreviousPlan, err := json.Marshal(previousPlan)
	if err != nil {
		return errs(err)
	}

	stdout, stderr, exitCode, err := a.CommandRunner.Run(a.ExternalBinPath, "generate-manifest", string(serialisedServiceDeployment),
		string(serialisedPlan), string(serialisedRequestParams), string(previousManifest), string(serialisedPreviousPlan))

	e := newErrorFactory(a.ExternalBinPath, stdout, stderr)
	if err != nil {
		return errs(e.adapterError(err))
	}

	switch *exitCode {
	case SuccessExitCode:
		break
	case serviceadapter.NotImplementedExitCode:
		return errs(e.exitCodeError(fmt.Errorf("service adapter has not implemented generate-manifest command: %s", err), *exitCode))
	default:
		return errs(e.exitCodeError(err, *exitCode))
	}

	logger.Printf("service adapter ran generate-manifest successfully, stderr logs: %s", string(stderr))

	validator := manifestValidator{deploymentName: serviceDeployment.DeploymentName}
	if err := validator.validateManifest(e, stdout); err != nil {
		return errs(err)
	}

	return stdout, nil
}

func (v manifestValidator) validateManifest(e errorFactory, stdout []byte) error {
	var generatedManifest manifest

	if err := yaml.Unmarshal(stdout, &generatedManifest); err != nil {
		return e.invalidYamlError()
	}

	if generatedManifest.Name != v.deploymentName {
		return e.incorrectDeploymentNameError(v.deploymentName, generatedManifest.Name)
	}

	for _, release := range generatedManifest.Releases {
		if strings.HasSuffix(release.Version, "latest") {
			return e.invalidVersionError(release.Version)
		}
	}

	for _, stemcell := range generatedManifest.Stemcells {
		if strings.HasSuffix(stemcell.Version, "latest") {
			return e.invalidVersionError(stemcell.Version)
		}
	}

	return nil
}
