package boshclient_test

import (
	"errors"

	"gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("deploying a manifest", func() {
	const (
		taskID = 3
	)

	var (
		deploymentManifest = bosh.BoshManifest{
			Name:           "big deployment",
			Releases:       []bosh.Release{},
			Stemcells:      []bosh.Stemcell{},
			InstanceGroups: []bosh.InstanceGroup{},
		}
		deploymentManifestBytes []byte

		returnedTaskID int
		deployErr      error
	)

	BeforeEach(func() {
		var err error
		deploymentManifestBytes, err = yaml.Marshal(deploymentManifest)
		Expect(err).NotTo(HaveOccurred())
	})

	Context("with a struct", func() {
		JustBeforeEach(func() {
			returnedTaskID, deployErr = c.Deploy(deploymentManifest, logger)
		})

		Context("and the deploy succeeds", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deploy().WithRawManifest(deploymentManifestBytes).WithoutContextID().RedirectsToTask(taskID),
				)
			})

			It("calls the authorization header builder", func() {
				Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})

			It("does not return an error", func() {
				Expect(deployErr).NotTo(HaveOccurred())
			})
		})

		Context("and the bosh director responds with an error", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deploy().WithoutContextID().Fails("because reasons"),
				)
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
			})
		})

		Context("and the bosh director cannot be reached", func() {
			BeforeEach(func() {
				director.Close()
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("error reaching bosh director")))
			})
		})

		Context("and the authorization header cannot be built", func() {
			BeforeEach(func() {
				authHeaderBuilder.BuildReturns("", errors.New("some-error"))
			})

			It("returns an error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("some-error")))
			})
		})
	})

	Context("with raw bytes", func() {
		var boshContextID string

		BeforeEach(func() {
			boshContextID = ""
		})

		JustBeforeEach(func() {
			returnedTaskID, deployErr = c.DeployRaw(deploymentManifestBytes, boshContextID, logger)
		})

		Context("and no bosh context id", func() {
			Context("and the deploy succeeds", func() {
				BeforeEach(func() {
					director.VerifyAndMock(
						mockbosh.Deploy().WithRawManifest(deploymentManifestBytes).WithoutContextID().RedirectsToTask(taskID),
					)
				})

				It("calls the authorization header builder", func() {
					Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
				})

				It("returns the bosh task ID", func() {
					Expect(returnedTaskID).To(Equal(taskID))
				})

				It("does not return an error", func() {
					Expect(deployErr).NotTo(HaveOccurred())
				})
			})
		})

		Context("and a bosh context id", func() {
			BeforeEach(func() {
				boshContextID = "bosh-context-id"

				director.VerifyAndMock(
					mockbosh.Deploy().WithRawManifest(deploymentManifestBytes).WithContextID("bosh-context-id").RedirectsToTask(taskID),
				)
			})

			It("calls the authorization header builder", func() {
				Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
			})

			It("returns the bosh task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})

			It("does not return an error", func() {
				Expect(deployErr).NotTo(HaveOccurred())
			})
		})

		Context("and the authorization header cannot be built", func() {
			BeforeEach(func() {
				authHeaderBuilder.BuildReturns("", errors.New("some-error"))
			})

			It("returns an error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("some-error")))
			})
		})

		Context("and the bosh director cannot be reached", func() {
			BeforeEach(func() {
				director.Close()
			})

			It("returns an error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("error reaching bosh director")))
			})
		})

		Context("and the bosh director responds with an error", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					mockbosh.Deploy().WithoutContextID().Fails("because reasons"),
				)
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
			})
		})
	})
})
