package boshclient

import (
	"fmt"
	"log"
	"net/http"
	"strconv"
	"strings"
)

const (
	semiSemverVersionLength = 2
	semverVersionLength     = 3
	stemcellVersionLength   = 4
)

func (c *Client) GetDirectorVersion(logger *log.Logger) (BoshDirectorVersion, error) {
	var boshInfo BoshInfo

	err := c.getDataFromBoshCheckingForErrors(fmt.Sprintf("%s/info", c.boshURL), http.StatusOK, &boshInfo, logger)
	if err != nil {
		return BoshDirectorVersion{}, err
	}

	version, err := newBoshDirectorVersion(boshInfo.Version)
	if err != nil {
		return BoshDirectorVersion{}, err
	}

	return version, nil
}

func newBoshDirectorVersion(rawVersion string) (BoshDirectorVersion, error) {
	trimmedVersion := strings.Fields(rawVersion)
	if len(trimmedVersion) == 0 {
		return BoshDirectorVersion{}, NewUnrecognisedBoshDirectorVersionError(rawVersion)
	}

	versionPart := trimmedVersion[0]
	versionNumbers := strings.Split(versionPart, ".")

	var versionType BoshDirectorVersionType

	switch len(versionNumbers) {
	case semiSemverVersionLength, semverVersionLength:
		versionType = SemverDirectorVersionType
	case stemcellVersionLength:
		versionType = StemcellDirectorVersionType
		versionNumbers = versionNumbers[1:4]
	default:
		return BoshDirectorVersion{}, NewUnrecognisedBoshDirectorVersionError(rawVersion)
	}

	majorVersion, err := strconv.Atoi(versionNumbers[0])
	if err != nil {
		return BoshDirectorVersion{}, NewUnrecognisedBoshDirectorVersionError(rawVersion)
	}

	return BoshDirectorVersion{majorVersion: majorVersion, versionType: versionType}, nil
}
