package broker

import (
	"context"
	"fmt"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

func (b *Broker) Bind(
	ctx context.Context,
	instanceID,
	bindingID string,
	details brokerapi.BindDetails,
) (brokerapi.Binding, error) {
	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeBind), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	errs := func(err DisplayableError) (brokerapi.Binding, error) {
		logger.Println(err)
		return brokerapi.Binding{}, err.ErrorForCFUser()
	}

	vms, manifest, err := b.getDeploymentInfo(instanceID, logger)
	switch err.(type) {
	case boshclient.DeploymentNotFoundError:
		return errs(NewDisplayableError(brokerapi.ErrInstanceDoesNotExist, fmt.Errorf("error binding: instance %s, not found", instanceID)))
	case error:
		return errs(
			NewGenericError(ctx, fmt.Errorf("gathering binding info %s", err)),
		)
	}

	logger.Printf("service adapter will create binding with ID %s for instance %s\n", bindingID, instanceID)
	mappedParams, err := convertToMap(details)
	if err != nil {
		return errs(NewGenericError(ctx, fmt.Errorf("converting to map %s", err)))
	}

	adapterBinding, err := b.serviceAdapter.CreateBinding(bindingID, vms, manifest, mappedParams, logger)

	switch err := err.(type) {
	case serviceadapter.BindingAlreadyExistsError:
		return errs(NewDisplayableError(brokerapi.ErrBindingAlreadyExists, fmt.Errorf("creating binding: %s", err)))
	case serviceadapter.AppGuidNotProvidedError:
		return errs(NewDisplayableError(brokerapi.ErrAppGuidNotProvided, fmt.Errorf("creating binding: %s", err)))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	if b.credentialStore.Enabled {
		credentialId := fmt.Sprintf("%s/%s", instanceID, bindingID)
		err := b.credentialStore.PutCredentials(credentialId, adapterBinding.Credentials)
		if err != nil {
			logger.Printf("Unable to put %s in credhub: %s\n", credentialId, err)
		}
	}

	return brokerapi.Binding{
		Credentials:     adapterBinding.Credentials,
		SyslogDrainURL:  adapterBinding.SyslogDrainURL,
		RouteServiceURL: adapterBinding.RouteServiceURL,
	}, nil
}
