package broker

import (
	"errors"
	"fmt"
	"log"
)

type deployer struct {
	boshClient        BoshClient
	manifestGenerator ManifestGenerator
	featureFlags      FeatureFlags
}

func NewDeployer(
	boshClient BoshClient,
	manifestGenerator ManifestGenerator,
	featureFlags FeatureFlags,
) Deployer {
	return deployer{
		boshClient:        boshClient,
		manifestGenerator: manifestGenerator,
		featureFlags:      featureFlags,
	}
}

func (d deployer) Deploy(instanceID, planID, deploymentName string, operationType OperationType, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string, boshContextID string, logger *log.Logger) (int, []byte, error) {
	err := d.assertNoOperationsInProgress(deploymentName, logger)
	if err != nil {
		return 0, nil, err
	}

	err = d.assertValidRequest(requestParams, planID, previousPlanID)
	if err != nil {
		return 0, nil, err
	}

	applyChanges := deleteApplyChanges(requestParams)

	if operationType == OperationTypeUpdate && !applyChanges {
		pendingChanges, pendingChangesErr := d.hasPendingChanges(deploymentName, instanceID, previousPlanID, operationType, oldManifest, logger)
		if pendingChangesErr != nil {
			return 0, nil, pendingChangesErr
		}

		if pendingChanges == true {
			err = errors.New("pending changes detected")
			if d.featureFlags.CFUserTriggeredUpgrades() {
				return 0, nil, NewPendingChangesError(err)
			} else {
				return 0, nil, NewApplyChangesDisabledError(err)
			}
		}
	}

	manifest, err := d.manifestGenerator.GenerateManifest(deploymentName, instanceID, planID, requestParams, oldManifest, previousPlanID, logger)
	if err != nil {
		return 0, nil, err
	}

	taskID, err := d.boshClient.DeployRaw(manifest, boshContextID, logger)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instanceID, taskID)

	return taskID, manifest, nil
}

func (d deployer) assertNoOperationsInProgress(deploymentName string, logger *log.Logger) error {
	tasks, err := d.boshClient.GetTasks(deploymentName, logger)
	if err != nil {
		return fmt.Errorf("error getting tasks for deployment %s: %s\n", deploymentName, err)
	}

	if incompleteTasks := tasks.IncompleteTasks(); len(incompleteTasks) != 0 {
		userError := errors.New("An operation is in progress for your service instance. Please try again later.")
		operatorError := NewOperationInProgressError(
			fmt.Errorf("deployment %s is still in progress: tasks %s\n",
				deploymentName,
				incompleteTasks.ToLog()),
		)
		return NewDisplayableError(userError, operatorError)
	}

	return nil
}

func (d deployer) assertValidRequest(requestParams map[string]interface{}, planID string, previousPlanID *string) error {
	parameters, ok := requestParams["parameters"].(map[string]interface{})
	if !ok {
		return nil
	}

	return d.assertValidApplyChangesRequest(parameters, planID, previousPlanID)
}

func (d deployer) assertValidApplyChangesRequest(parameters map[string]interface{}, planID string, previousPlanID *string) error {
	if parameters["apply-changes"] == nil {
		return nil
	}

	applyChanges, ok := parameters["apply-changes"].(bool)
	if !ok {
		err := errors.New("update called with apply-changes set to non-boolean")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return NewPendingChangesError(err)
		} else {
			return NewApplyChangesNotPermittedError(err)
		}
	}

	if !applyChanges {
		return nil
	}

	if !d.featureFlags.CFUserTriggeredUpgrades() {
		return NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))
	}

	if previousPlanID != nil && planID != *previousPlanID {
		err := errors.New("update called with apply-changes and a plan change")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return NewPendingChangesError(err)
		} else {
			return NewApplyChangesNotPermittedError(err)
		}
	}

	if len(parameters) > 1 {
		err := errors.New("update called with apply-changes and arbitrary parameters set")
		if d.featureFlags.CFUserTriggeredUpgrades() {
			return NewPendingChangesError(err)
		} else {
			return NewApplyChangesNotPermittedError(err)
		}
	}

	return nil
}

func deleteApplyChanges(requestParams map[string]interface{}) bool {
	parameters, ok := requestParams["parameters"].(map[string]interface{})
	if !ok {
		return false
	}

	applyChanges, ok := parameters["apply-changes"].(bool)
	if !ok {
		return false
	}

	delete(parameters, "apply-changes")

	return applyChanges
}

func (d deployer) hasPendingChanges(deploymentName, instanceID string, previousPlanID *string, operationType OperationType, oldManifest BoshManifest, logger *log.Logger) (bool, error) {
	// generate a manifest without params and the previously used plan ID
	regeneratedManifest, err := d.manifestGenerator.GenerateManifest(deploymentName, instanceID, *previousPlanID, map[string]interface{}{}, oldManifest, previousPlanID, logger)
	if err != nil {
		return false, err
	}

	// check whether the manifest generated by the adapter is the same we previously deployed
	manifestsEqual, err := regeneratedManifest.Equals(oldManifest)
	if err != nil {
		return false, fmt.Errorf("error detecting change in manifest: %s", err)
	}

	return !manifestsEqual, nil
}
