package broker_test

import (
	"errors"
	"fmt"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/broker/fakes"
)

var _ = Describe("Deployer", func() {
	const (
		instanceID     = "some-instance-id"
		deploymentName = "a-fancy-deployment"
		boshTaskID     = 42
	)

	var (
		deployer         broker.Deployer
		planID           string
		previousPlanID   *string
		boshContextID    string
		requestParams    map[string]interface{}
		oldManifest      []byte
		deployedManifest []byte
		deployError      error
		operationType    broker.OperationType
		returnedTaskID   int

		manifest = []byte("---\nmanifest: thing")

		manifestGenerator *fakes.FakeManifestGenerator
		featureFlags      *fakes.FakeFeatureFlags
	)

	BeforeEach(func() {
		manifestGenerator = new(fakes.FakeManifestGenerator)
		featureFlags = new(fakes.FakeFeatureFlags)
		featureFlags.CFUserTriggeredUpgradesReturns(false)

		deployer = broker.NewDeployer(boshClient, manifestGenerator, featureFlags)

		//defaults
		planID = existingPlanID
		previousPlanID = nil
		requestParams = map[string]interface{}{}
		oldManifest = nil
		operationType = ""
		boshContextID = ""
	})

	JustBeforeEach(func() {
		logger := loggerFactory.NewLogger()
		returnedTaskID, deployedManifest, deployError = deployer.Deploy(
			instanceID,
			planID,
			deploymentName,
			operationType,
			requestParams,
			oldManifest,
			previousPlanID,
			boshContextID,
			logger,
		)
	})

	Context("when creating a deployment", func() {
		BeforeEach(func() {
			oldManifest = nil
			previousPlanID = nil
			operationType = broker.OperationTypeCreate
		})

		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				By("not having any previous tasks")
				boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
				manifestGenerator.GenerateManifestReturns(manifest, nil)
				boshClient.DeployRawReturns(42, nil)
			})

			It("checks tasks for the deployment", func() {
				Expect(boshClient.GetTasksCallCount()).To(Equal(1))
				actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
				Expect(actualDeploymentName).To(Equal(deploymentName))
			})

			It("calls generate manifest", func() {
				Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(1))
			})

			It("calls new manifest with correct params", func() {
				Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(1))
				passedDeploymentName, passedInstanceID, passedPlanID, passedRequestParams, passedPreviousManifest, passedPreviousPlanID, _ := manifestGenerator.GenerateManifestArgsForCall(0)

				Expect(passedDeploymentName).To(Equal(deploymentName))
				Expect(passedInstanceID).To(Equal(instanceID))
				Expect(passedPlanID).To(Equal(planID))
				Expect(passedRequestParams).To(Equal(requestParams))
				Expect(passedPreviousManifest).To(Equal(oldManifest))
				Expect(passedPreviousPlanID).To(Equal(previousPlanID))
			})

			It("returns the bosh task ID", func() {
				Expect(returnedTaskID).To(Equal(boshTaskID))
			})

			It("Creates a bosh deployment using generated manifest", func() {
				Expect(boshClient.DeployRawCallCount()).To(Equal(1))
				deployedManifest, _, _ := boshClient.DeployRawArgsForCall(0)
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("return the newly generated manifest", func() {
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("does not return an error", func() {
				Expect(deployError).NotTo(HaveOccurred())
			})

			Context("when bosh context ID is provided", func() {
				BeforeEach(func() {
					boshContextID = "bosh-context-id"
				})

				It("invokes boshclient's DeployRaw with context ID", func() {
					Expect(boshClient.DeployRawCallCount()).To(Equal(1))
					_, actualBoshContextID, _ := boshClient.DeployRawArgsForCall(0)
					Expect(actualBoshContextID).To(Equal(boshContextID))
				})
			})
		})
	})

	Context("logging", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(42, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
		})

		Context("for creating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeCreate
				oldManifest = nil
				manifestGenerator.GenerateManifestReturns(manifest, nil)
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(deployError).ToNot(HaveOccurred())
				Expect(returnedTaskID).To(Equal(42))
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for create instance %s is %d", instanceID, boshTaskID)))
			})
		})

		Context("for updating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeUpdate
				previousPlanID = stringPointer(secondPlanID)
				oldManifest = []byte(`---\nfoo: bar`)
				manifestGenerator.GenerateManifestReturns(oldManifest, nil)
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(deployError).ToNot(HaveOccurred())
				Expect(returnedTaskID).To(Equal(42))
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for update instance %s is %d", instanceID, boshTaskID)))
			})
		})

		Context("for upgrading", func() {
			BeforeEach(func() {
				manifestGenerator.GenerateManifestReturns(manifest, nil)
				operationType = broker.OperationTypeUpgrade
				previousPlanID = stringPointer(secondPlanID)
			})

			It("logs the bosh task ID returned by the director", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for upgrade instance %s is %d", instanceID, boshTaskID)))
			})
		})
	})

	Context("when updating a deployment", func() {
		BeforeEach(func() {
			oldManifest = []byte("---\nold-manifest-fetched-from-bosh: bar")
			operationType = broker.OperationTypeUpdate
			previousPlanID = stringPointer(existingPlanID)
			delete(secondPlan.Properties, "a_global_property") // test that global properties actually get merged by removing overriding plan property in this context

			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
		})

		Context("and the manifest generator fails to generate the manifest the first time", func() {
			BeforeEach(func() {
				manifestGenerator.GenerateManifestReturns(nil, errors.New("manifest fail"))
			})

			It("wraps the error", func() {
				Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
			})
		})

		Context("and there are no pending changes", func() {
			Context("and manifest generation succeeds", func() {
				BeforeEach(func() {
					requestParams = map[string]interface{}{"foo": "bar"}
					manifestGenerator.GenerateManifestStub = func(
						_, _, _ string,
						requestParams map[string]interface{},
						previousManifest []byte,
						_ *string,
						_ *log.Logger,
					) (broker.BoshManifest, error) {
						if len(requestParams) > 0 {
							return manifest, nil
						}
						return previousManifest, nil
					}

					boshClient.DeployRawReturns(42, nil)
				})

				It("checks tasks for the deployment", func() {
					Expect(boshClient.GetTasksCallCount()).To(Equal(1))
					actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
					Expect(actualDeploymentName).To(Equal(deploymentName))
				})

				It("generate manifest without arbitrary params", func() {
					Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(2))
					_, _, _, passedRequestParams, _, _, _ := manifestGenerator.GenerateManifestArgsForCall(0)
					Expect(passedRequestParams).To(BeEmpty())
				})

				It("generates new manifest with arbitrary params", func() {
					Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(2))
					_, _, _, passedRequestParams, _, _, _ := manifestGenerator.GenerateManifestArgsForCall(1)
					Expect(passedRequestParams).To(Equal(requestParams))
				})

				It("Creates a bosh deployment from the generated manifest", func() {
					Expect(boshClient.DeployRawCallCount()).To(Equal(1))
					deployedManifest, _, _ := boshClient.DeployRawArgsForCall(0)
					Expect(string(deployedManifest)).To(Equal(string(manifest)))
				})

				It("returns the bosh task ID", func() {
					Expect(returnedTaskID).To(Equal(boshTaskID))
				})
			})

			Context("and the manifest generator fails to generate the manifest the second time", func() {
				BeforeEach(func() {
					requestParams = map[string]interface{}{"foo": "bar"}
					manifestGenerator.GenerateManifestStub = func(
						_, _, _ string,
						requestParams map[string]interface{},
						previousManifest []byte,
						_ *string,
						_ *log.Logger,
					) (broker.BoshManifest, error) {
						if len(requestParams) > 0 {
							return nil, errors.New("manifest fail")
						}
						return previousManifest, nil
					}
				})

				It("wraps the error", func() {
					Expect(boshClient.DeployRawCallCount()).To(Equal(0))
					Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
				})
			})
		})

		Context("and there are pending changes", func() {
			BeforeEach(func() {
				manifestGenerator.GenerateManifestReturns(manifest, nil)
			})

			Context("and apply-changes param is set to true", func() {
				BeforeEach(func() {
					requestParams["parameters"] = map[string]interface{}{"apply-changes": true}
				})

				Context("and there are no other params set", func() {
					Context("and the cf_user_triggered_upgrades feature is turned on", func() {
						BeforeEach(func() {
							featureFlags.CFUserTriggeredUpgradesReturns(true)
						})

						It("checks if the feature is on", func() {
							Expect(featureFlags.CFUserTriggeredUpgradesCallCount()).To(Equal(1))
						})

						It("does deploy the manifest again", func() {
							Expect(boshClient.DeployRawCallCount()).To(Equal(1))
							manifestBytes, _, _ := boshClient.DeployRawArgsForCall(0)
							Expect(manifestBytes).To(Equal(manifest))
						})

						It("does not return an error", func() {
							Expect(deployError).ToNot(HaveOccurred())
						})

						It("does not pass apply-changes to generate updated manifest", func() {
							_, _, _, passedRequestParams, _, _, _ := manifestGenerator.GenerateManifestArgsForCall(0)
							Expect(passedRequestParams["parameters"]).To(Equal(map[string]interface{}{}))
						})
					})

					Context("and the cf_user_triggered_upgrades feature is disabled", func() {
						It("does not deploy the manifest again", func() {
							Expect(boshClient.DeployRawCallCount()).To(Equal(0))
						})

						It("does not call generate manifest", func() {
							Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(0))
						})

						It("returns an apply changes not permitted error", func() {
							Expect(deployError).To(Equal(broker.NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))))
						})
					})
				})

				Context("and there is a plan change", func() {
					BeforeEach(func() {
						previousPlanID = stringPointer(secondPlanID)
					})

					It("checks if the feature is on", func() {
						Expect(featureFlags.CFUserTriggeredUpgradesCallCount()).To(Equal(1))
					})

					It("does not call generate manifest", func() {
						Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(0))
					})

					It("does NOT deploy", func() {
						Expect(boshClient.DeployRawCallCount()).To(BeZero())
					})

					Context("and the cf_user_triggered_upgrades feature is turned on", func() {
						BeforeEach(func() {
							featureFlags.CFUserTriggeredUpgradesReturns(true)
						})

						It("returns an pending changes error", func() {
							Expect(deployError).To(Equal(broker.NewPendingChangesError(errors.New("update called with apply-changes and a plan change"))))
						})
					})

					Context("and the cf_user_triggered_upgrades feature is disabled", func() {
						It("returns an apply changes disabled error", func() {
							Expect(deployError).To(Equal(broker.NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))))
						})
					})
				})

				Context("and there are other params set", func() {
					BeforeEach(func() {
						requestParams["parameters"] = map[string]interface{}{"apply-changes": true, "foo": "bar"}
					})

					It("does not call generate manifest", func() {
						Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(0))
					})

					It("does NOT deploy", func() {
						Expect(boshClient.DeployRawCallCount()).To(BeZero())
					})

					Context("and the cf_user_triggered_upgrades feature is turned on", func() {
						BeforeEach(func() {
							featureFlags.CFUserTriggeredUpgradesReturns(true)
						})

						It("returns a pending changes error", func() {
							Expect(deployError).To(Equal(broker.NewPendingChangesError(errors.New("update called with apply-changes and arbitrary parameters set"))))
						})
					})

					Context("and the cf_user_triggered_upgrades feature is disabled", func() {
						It("returns an apply changes not permitted error", func() {
							Expect(deployError).To(Equal(broker.NewApplyChangesNotPermittedError(errors.New("'cf_user_triggered_upgrades' feature is disabled"))))
						})
					})
				})
			})

			Context("and apply-changes param is set to false", func() {
				BeforeEach(func() {
					requestParams["parameters"] = map[string]interface{}{"apply-changes": false}
				})

				It("does call generate manifest once", func() {
					Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(1))
				})

				It("does NOT deploy", func() {
					Expect(boshClient.DeployRawCallCount()).To(BeZero())
				})

				Context("and the cf_user_triggered_upgrades feature is turned on", func() {
					BeforeEach(func() {
						featureFlags.CFUserTriggeredUpgradesReturns(true)
					})

					It("checks if the feature is on", func() {
						Expect(featureFlags.CFUserTriggeredUpgradesCallCount()).To(Equal(1))
					})

					It("returns a pending changes error", func() {
						Expect(deployError).To(Equal(broker.NewPendingChangesError(errors.New("pending changes detected"))))
					})
				})

				Context("and the cf_user_triggered_upgrades feature is disabled", func() {
					It("returns an apply changes disabled error", func() {
						Expect(deployError).To(Equal(broker.NewApplyChangesDisabledError(errors.New("pending changes detected"))))
					})
				})
			})

			Context("and apply-changes param is not set at all", func() {
				It("checks tasks for the deployment", func() {
					Expect(boshClient.GetTasksCallCount()).To(Equal(1))
					actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
					Expect(actualDeploymentName).To(Equal(deploymentName))
				})

				It("generate manifest without arbitrary params", func() {
					Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(1))
					_, _, _, passedRequestParams, _, _, _ := manifestGenerator.GenerateManifestArgsForCall(0)
					Expect(passedRequestParams).To(BeEmpty())
				})

				It("does not deploy", func() {
					Expect(boshClient.DeployRawCallCount()).To(Equal(0))
				})

				Context("and the cf_user_triggered_upgrades feature is turned on", func() {
					BeforeEach(func() {
						featureFlags.CFUserTriggeredUpgradesReturns(true)
					})

					It("checks if the feature is on", func() {
						Expect(featureFlags.CFUserTriggeredUpgradesCallCount()).To(Equal(1))
					})

					It("returns a pending changes error", func() {
						Expect(deployError).To(Equal(broker.NewPendingChangesError(errors.New("pending changes detected"))))
					})
				})

				Context("and the cf_user_triggered_upgrades feature is disabled", func() {
					It("returns an apply changes disabled error", func() {
						Expect(deployError).To(Equal(broker.NewApplyChangesDisabledError(errors.New("pending changes detected"))))
					})
				})

				Context("when an error occurs detecting a pending change", func() {
					BeforeEach(func() {
						invalidManifest := []byte(`this isnt yaml!`)
						manifestGenerator.GenerateManifestReturns(invalidManifest, nil)
					})

					It("wraps the error", func() {
						Expect(deployError).To(MatchError(ContainSubstring("error detecting change in manifest: yaml: unmarshal errors:")))
					})
				})
			})

			Context("and apply-changes param is set to non-boolean", func() {
				BeforeEach(func() {
					requestParams["parameters"] = map[string]interface{}{"apply-changes": 42}
				})

				It("does NOT call generate manifest", func() {
					Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(0))
				})

				It("does NOT deploy", func() {
					Expect(boshClient.DeployRawCallCount()).To(BeZero())
				})

				Context("and the cf_user_triggered_upgrades feature is turned on", func() {
					BeforeEach(func() {
						featureFlags.CFUserTriggeredUpgradesReturns(true)
					})

					It("checks if the feature is on", func() {
						Expect(featureFlags.CFUserTriggeredUpgradesCallCount()).To(Equal(1))
					})

					It("returns a pending changes error", func() {
						Expect(deployError).To(Equal(broker.NewPendingChangesError(errors.New("update called with apply-changes set to non-boolean"))))
					})
				})

				Context("and the cf_user_triggered_upgrades feature is disabled", func() {
					It("returns an apply changes disabled error", func() {
						Expect(deployError).To(Equal(broker.NewApplyChangesNotPermittedError(errors.New("update called with apply-changes set to non-boolean"))))
					})
				})
			})
		})
	})

	Context("when upgrading a deployment", func() {
		BeforeEach(func() {
			oldManifest = []byte("---\nold-manifest-fetched-from-bosh: bar")
			operationType = broker.OperationTypeUpgrade
			previousPlanID = stringPointer(secondPlanID)

			boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
			manifestGenerator.GenerateManifestReturns(manifest, nil)
			boshClient.DeployRawReturns(42, nil)
		})

		It("does not check for pending changes", func() {
			Expect(manifestGenerator.GenerateManifestCallCount()).To(Equal(1))
		})

		It("returns the bosh task ID and new manifest", func() {
			Expect(returnedTaskID).To(Equal(42))
			Expect(deployedManifest).To(Equal(manifest))
			Expect(deployError).NotTo(HaveOccurred())
		})
	})

	Context("when manifest generator returns an error", func() {
		BeforeEach(func() {
			manifestGenerator.GenerateManifestReturns(nil, fmt.Errorf("some error"))
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			requestParams = map[string]interface{}{"foo": "bar"}
			operationType = broker.OperationTypeCreate
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			actualDeploymentName, _ := boshClient.GetTasksArgsForCall(0)
			Expect(actualDeploymentName).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("some error")))
		})
	})

	Context("when the last bosh task for deployment is queued", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskQueued, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
			operationType = broker.OperationTypeCreate
		})

		It("returns an error for the operator", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(MatchError(ContainSubstring("{\"ID\":%d", boshTaskID)))
		})

		It("returns an error for the CF user", func() {
			Expect(deployError).To(BeAssignableToTypeOf(broker.DisplayableError{}))
			displayableErr, _ := deployError.(broker.DisplayableError)
			Expect(displayableErr.ErrorForCFUser()).To(MatchError("An operation is in progress for your service instance. Please try again later."))
		})

		It("does not log the previous completed tasks for the deployment", func() {
			Expect(logBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("\"ID\":%d", previousDoneBoshTaskID))
			Expect(logBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("\"ID\":%d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment is processing", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskProcessing, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
			operationType = broker.OperationTypeCreate
		})

		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(MatchError((ContainSubstring("\"ID\":%d", boshTaskID))))
		})

		It("returns an error for the CF user", func() {
			Expect(deployError).To(BeAssignableToTypeOf(broker.DisplayableError{}))
			displayableErr, _ := deployError.(broker.DisplayableError)
			Expect(displayableErr.ErrorForCFUser()).To(MatchError("An operation is in progress for your service instance. Please try again later."))
		})

		It("does not log the previous tasks for the deployment", func() {
			Expect(logBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("\"ID\":%d", previousDoneBoshTaskID))
			Expect(logBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(logBuffer.String()).NotTo(ContainSubstring("\"ID\":%d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment fails to fetch", func() {
		BeforeEach(func() {
			operationType = broker.OperationTypeCreate
			boshClient.GetTasksReturns(nil, fmt.Errorf("connection error"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(fmt.Sprintf("error getting tasks for deployment %s: connection error\n", deploymentName)))
		})
	})

	Context("when bosh fails to deploy the release", func() {
		BeforeEach(func() {
			operationType = broker.OperationTypeCreate
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			boshClient.DeployRawReturns(0, errors.New("oops"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("oops")))
		})
	})
})

func stringPointer(s string) *string {
	return &s
}
