package broker_test

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/bosh"
)

var _ = Describe("deprovisioning instances", func() {
	var (
		instanceID = "an-instance-to-be-deprovisioned"

		asyncAllowed    bool
		deprovisionSpec brokerapi.DeprovisionServiceSpec
		deprovisionErr  error

		deleteTaskID = 88
		boshVMs      = bosh.BoshVMs{"instance-group": {"ip.from.bosh"}}
		boshManifest = []byte("name: service-instance_an-instance-to-be-deprovisioned")
	)

	BeforeEach(func() {
		asyncAllowed = true
		boshClient.VMsReturns(boshVMs, nil)
		boshClient.GetManifestRawReturns(boshManifest, nil)
		boshClient.DeleteDeploymentReturns(deleteTaskID, nil)
	})

	JustBeforeEach(func() {
		deprovisionSpec, deprovisionErr = b.Deprovision(context.Background(), instanceID, brokerapi.DeprovisionDetails{}, asyncAllowed)
	})

	It("returns that is deprovisioning asynchronously", func() {
		Expect(deprovisionSpec.IsAsync).To(BeTrue())
	})

	It("returns the bosh task ID and operation type in the operation data", func() {
		var operationData broker.OperationData
		Expect(json.Unmarshal([]byte(deprovisionSpec.OperationData), &operationData)).To(Succeed())
		Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deleteTaskID, OperationType: broker.OperationTypeDelete}))
	})

	It("deletes a bosh deployment whose name is based on the instance ID", func() {
		Expect(boshClient.DeleteDeploymentCallCount()).To(Equal(1))
		actualInstanceID, _ := boshClient.DeleteDeploymentArgsForCall(0)
		Expect(actualInstanceID).To(Equal(deploymentName(instanceID)))
	})

	It("logs that it will delete the deployment with a request ID", func() {
		Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} deleting deployment for instance`))
	})

	It("logs the task id to delete the deployment", func() {
		Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Delete instance %s was %d", instanceID, deleteTaskID)))
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})

		It("returns an error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Describe("the deployment not existing", func() {
		Context("when deleting the deployment", func() {
			BeforeEach(func() {
				boshClient.DeleteDeploymentReturns(0, boshclient.DeploymentNotFoundError{})
			})

			It("returns an error", func() {
				Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error deprovisioning: instance %s, not found", instanceID)))
			})
		})
	})

	Describe("when deleteing a deployment fails", func() {
		Context("with a generic error", func() {
			BeforeEach(func() {
				boshClient.DeleteDeploymentReturns(0, errors.New("er ma gerd!"))
			})

			Describe("returned error", func() {
				It("includes a standard message", func() {
					Expect(deprovisionErr).To(MatchError(ContainSubstring(
						"There was a problem completing your request. Please contact your operations team providing the following information:",
					)))
				})

				It("includes the broker request id", func() {
					Expect(deprovisionErr).To(MatchError(MatchRegexp(
						`broker-request-id: [0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}`,
					)))
				})

				It("includes the service name", func() {
					Expect(deprovisionErr).To(MatchError(ContainSubstring(
						"service: a-cool-redis-service",
					)))
				})

				It("includes the service instance guid", func() {
					Expect(deprovisionErr).To(MatchError(ContainSubstring(
						fmt.Sprintf("service-instance-guid: %s", instanceID),
					)))
				})

				It("includes the operation type", func() {
					Expect(deprovisionErr).To(MatchError(ContainSubstring(
						"operation: delete",
					)))
				})

				It("does NOT include the bosh task id", func() {
					Expect(deprovisionErr).NotTo(MatchError(ContainSubstring(
						"task-id:",
					)))
				})
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
			})
		})

		Context("with a bosh request error", func() {
			BeforeEach(func() {
				boshClient.DeleteDeploymentReturns(0, boshclient.NewRequestError(
					fmt.Errorf("error deleting instance: network timeout"),
				))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error deleting instance: network timeout"))
			})

			It("returns the try again later error for the user", func() {
				Expect(deprovisionErr).To(MatchError(ContainSubstring("Currently unable to delete service instance, please try again later")))
			})
		})
	})
})
