package broker

import (
	"context"
	"errors"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

const genericErrorPrefix = "There was a problem completing your request. Please contact your operations team providing the following information:"

type NotFoundError struct {
	error
}

func NewNotFoundError() NotFoundError {
	return NotFoundError{error: errors.New("resource was not found")}
}

func IsNotFound(err error) bool {
	switch err.(type) {
	case NotFoundError:
		return true
	default:
		return false
	}
}

type DeploymentNotFoundError struct {
	error
}

func NewDeploymentNotFoundError(e error) error {
	return DeploymentNotFoundError{e}
}

type OperationInProgressError struct {
	error
}

func NewOperationInProgressError(e error) error {
	return OperationInProgressError{e}
}

type ServiceAdapterCommandNotImplementedError struct {
	error
}

func NewServiceAdapterCommandNotImplementedError(e error) ServiceAdapterCommandNotImplementedError {
	return ServiceAdapterCommandNotImplementedError{e}
}

var NilError = DisplayableError{nil, nil}

type DisplayableError struct {
	errorForCFUser   error
	errorForOperator error
}

func (e DisplayableError) ErrorForCFUser() error {
	return e.errorForCFUser
}

func (e DisplayableError) Error() string {
	return fmt.Sprintf("error: %s. error for user: %s.", e.errorForOperator, e.errorForCFUser)
}

func (e DisplayableError) Occurred() bool {
	return e.errorForCFUser != nil && e.errorForOperator != nil
}

func NewDisplayableError(errorForCFUser, errForOperator error) DisplayableError {
	return DisplayableError{
		errorForCFUser,
		errForOperator,
	}
}

func NewBoshRequestError(action string, requestError error) DisplayableError {
	return DisplayableError{
		fmt.Errorf("Currently unable to %s service instance, please try again later", action),
		requestError,
	}
}

func NewPendingChangesError(errForOperator error) DisplayableError {
	return DisplayableError{
		errors.New(`There is a pending change to your service instance, you must first run cf update-service <service_name> -c '{"apply-changes": true}', no other arbitrary parameters are allowed`),
		errForOperator,
	}
}

func NewApplyChangesDisabledError(errForOperator error) DisplayableError {
	return DisplayableError{
		errors.New(`Service cannot be updated at this time, please try again later or contact your operator for more information`),
		errForOperator,
	}
}

func NewApplyChangesNotPermittedError(errForOperator error) DisplayableError {
	return DisplayableError{
		errors.New(`'apply-changes' is not permitted. Contact your operator for more information`),
		errForOperator,
	}
}

func NewGenericError(ctx context.Context, err error) DisplayableError {
	serviceName := brokercontext.GetServiceName(ctx)
	instanceID := brokercontext.GetInstanceID(ctx)
	reqID := brokercontext.GetReqID(ctx)
	operation := brokercontext.GetOperation(ctx)
	boshTaskID := brokercontext.GetBoshTaskID(ctx)

	message := fmt.Sprintf(
		"%s service: %s, service-instance-guid: %s, broker-request-id: %s",
		genericErrorPrefix,
		serviceName,
		instanceID,
		reqID,
	)

	if boshTaskID != 0 {
		message += fmt.Sprintf(", task-id: %d", boshTaskID)
	}

	if operation != "" {
		message += fmt.Sprintf(", operation: %s", operation)
	}

	return DisplayableError{
		errorForCFUser:   errors.New(message),
		errorForOperator: err,
	}
}
