package broker

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"log"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

var descriptions = map[brokerapi.LastOperationState]map[OperationType]string{
	brokerapi.InProgress: {
		OperationTypeCreate:  "Instance provisioning in progress",
		OperationTypeUpdate:  "Instance update in progress",
		OperationTypeUpgrade: "Instance upgrade in progress",
		OperationTypeDelete:  "Instance deletion in progress",
	},
	brokerapi.Succeeded: {
		OperationTypeCreate:  "Instance provisioning completed",
		OperationTypeUpdate:  "Instance update completed",
		OperationTypeUpgrade: "Instance upgrade completed",
		OperationTypeDelete:  "Instance deletion completed",
	},
	brokerapi.Failed: {
		OperationTypeCreate:  "Instance provisioning failed",
		OperationTypeUpdate:  "Instance update failed",
		OperationTypeUpgrade: "Failed for bosh task",
		OperationTypeDelete:  "Instance deletion failed",
	},
}

func (b *Broker) LastOperation(
	ctx context.Context,
	instanceID,
	operationDataRaw string,
) (brokerapi.LastOperation, error) {
	requestID := uuid.New()
	ctx = brokercontext.New(ctx, "", requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	errs := func(err DisplayableError) (brokerapi.LastOperation, error) {
		logger.Println(err)
		return brokerapi.LastOperation{}, err.ErrorForCFUser()
	}

	if operationDataRaw == "" {
		err := errors.New("Request missing operation data, please check your Cloud Foundry version is v238+")
		return errs(NewGenericError(ctx, err))
	}

	var operationData OperationData
	if err := json.Unmarshal([]byte(operationDataRaw), &operationData); err != nil {
		return errs(NewGenericError(
			ctx, fmt.Errorf("operation data cannot be parsed: %s", err),
		))
	}

	ctx = brokercontext.WithOperation(ctx, string(operationData.OperationType))

	if operationData.BoshTaskID == 0 {
		return errs(NewGenericError(
			ctx, errors.New("no task ID found in operation data"),
		))
	}

	ctx = brokercontext.WithBoshTaskID(ctx, operationData.BoshTaskID)

	lifeCycleRunner := NewLifeCycleRunner(b.boshClient, b.serviceOffering.Plans)

	lastBoshTask, err := lifeCycleRunner.GetTask(deploymentName(instanceID), operationData, logger)
	if err != nil {
		return errs(NewGenericError(ctx, fmt.Errorf(
			"error retrieving tasks from bosh, for deployment '%s': %s",
			deploymentName(instanceID), err,
		)))
	}

	ctx = brokercontext.WithBoshTaskID(ctx, lastBoshTask.ID)

	lastOperation := constructLastOperation(ctx, lastBoshTask, operationData, logger)
	logLastOperation(instanceID, lastBoshTask, operationData, logger)

	return lastOperation, nil
}

func constructLastOperation(ctx context.Context, boshTask boshclient.BoshTask, operationData OperationData, logger *log.Logger) brokerapi.LastOperation {
	taskState := lastOperationState(boshTask.State, logger)
	description := descriptionForOperationTask(ctx, taskState, operationData, boshTask.ID)

	return brokerapi.LastOperation{State: taskState, Description: description}
}

func lastOperationState(taskState string, logger *log.Logger) brokerapi.LastOperationState {
	switch taskState {
	case boshclient.BoshTaskQueued, boshclient.BoshTaskProcessing:
		return brokerapi.InProgress
	case boshclient.BoshTaskDone:
		return brokerapi.Succeeded
	case boshclient.BoshTaskError, boshclient.BoshTaskCancelled, boshclient.BoshTaskCancelling, boshclient.BoshTaskTimeout:
		return brokerapi.Failed
	default:
		logger.Printf("Unrecognised BOSH task state: %s", taskState)
		return brokerapi.Failed
	}
}

func descriptionForOperationTask(ctx context.Context, taskState brokerapi.LastOperationState, operationData OperationData, taskID int) string {
	description := descriptions[taskState][operationData.OperationType]

	if taskState == brokerapi.Failed {
		if operationData.OperationType == OperationTypeUpgrade {
			description = fmt.Sprintf(description+": %d", taskID) // Allows upgrader to log BOSH task ID when an upgrade fails
		} else {
			description = fmt.Sprintf(description+": %s", NewGenericError(ctx, nil).ErrorForCFUser())
		}
	}

	return description
}

func logLastOperation(instanceID string, boshTask boshclient.BoshTask, operationData OperationData, logger *log.Logger) {
	logger.Printf(
		"BOSH task ID %d status: %s %s deployment for instance %s: Description: %s Result: %s\n",
		boshTask.ID,
		boshTask.State,
		operationData.OperationType,
		instanceID,
		boshTask.Description,
		boshTask.Result,
	)
}
