package broker

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type LifeCycleRunner struct {
	boshClient BoshClient
	plans      Plans
}

func NewLifeCycleRunner(
	boshClient BoshClient,
	plans Plans,
) LifeCycleRunner {
	return LifeCycleRunner{
		boshClient,
		plans,
	}
}

func (l LifeCycleRunner) GetTask(deploymentName string, operationData OperationData, logger *log.Logger) (boshclient.BoshTask, error) {
	if operationData.BoshContextID == "" || !validPostDeployOpType(operationData.OperationType) {
		return l.boshClient.GetTask(operationData.BoshTaskID, logger)
	}

	boshTasks, err := l.boshClient.GetNormalisedTasksByContext(deploymentName, operationData.BoshContextID, logger)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	switch len(boshTasks) {
	case 0:
		return boshclient.BoshTask{}, fmt.Errorf("no tasks found for context id: %s", operationData.BoshContextID)
	case 1:
		task := boshTasks[0]

		if !task.Done() {
			return task, nil
		}

		plan, found := l.plans.FindByID(operationData.PlanID)
		if !found {
			logger.Printf("can't determine lifecycle errands, plan with id %s not found\n", operationData.PlanID)
			return task, nil
		}

		errand := l.postDeployErrand(plan)
		if errand == "" {
			return task, nil
		}

		return l.runErrand(deploymentName, errand, operationData.BoshContextID, logger)
	case 2:
		// there must be a post deploy errand and it must be the first in the task list
		return boshTasks[0], nil
	default:
		return boshclient.BoshTask{},
			fmt.Errorf("unexpected tasks found with context id: %s, tasks: %s", operationData.BoshContextID, boshTasks.ToLog())
	}
}

func validPostDeployOpType(op OperationType) bool {
	return op == OperationTypeCreate ||
		op == OperationTypeUpdate ||
		op == OperationTypeUpgrade
}

func (l LifeCycleRunner) postDeployErrand(plan Plan) string {
	if plan.LifecycleErrands == nil {
		return ""
	}

	return plan.LifecycleErrands.PostDeploy
}

func (l LifeCycleRunner) runErrand(deploymentName, errand, contextID string, log *log.Logger) (boshclient.BoshTask, error) {
	taskID, err := l.boshClient.RunErrand(deploymentName, errand, contextID, log)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	task, err := l.boshClient.GetTask(taskID, log)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	return task, nil
}
