package broker

import (
	"fmt"
	"log"
	"reflect"

	yaml "gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

type manifestGenerator struct {
	serviceAdapter  ServiceAdapterClient
	serviceOffering ServiceOffering
	serviceStemcell serviceadapter.Stemcell
	serviceReleases serviceadapter.ServiceReleases
}

func NewManifestGenerator(
	serviceAdapter ServiceAdapterClient,
	serviceOffering ServiceOffering,
	serviceStemcell serviceadapter.Stemcell,
	serviceReleases serviceadapter.ServiceReleases,
) manifestGenerator {
	return manifestGenerator{
		serviceAdapter:  serviceAdapter,
		serviceOffering: serviceOffering,
		serviceStemcell: serviceStemcell,
		serviceReleases: serviceReleases,
	}
}

type BoshManifest []byte

func (m BoshManifest) Equals(other BoshManifest) (bool, error) {
	var thisMap map[string]interface{}
	var thatMap map[string]interface{}

	err := yaml.Unmarshal(m, &thisMap)
	if err != nil {
		return false, err
	}
	err = yaml.Unmarshal(other, &thatMap)
	if err != nil {
		return false, err
	}

	return reflect.DeepEqual(thisMap, thatMap), nil
}

func (m manifestGenerator) GenerateManifest(
	deploymentName, instanceID, planID string,
	requestParams map[string]interface{},
	oldManifest []byte,
	previousPlanID *string,
	logger *log.Logger,
) (BoshManifest, error) {

	serviceDeployment := serviceadapter.ServiceDeployment{
		DeploymentName: deploymentName,
		Releases:       m.serviceReleases,
		Stemcell:       m.serviceStemcell,
	}

	plan, previousPlan, err := m.findPlans(planID, previousPlanID)
	if err != nil {
		return nil, err
	}

	logger.Printf("service adapter will generate manifest for instance %s\n", instanceID)

	return m.serviceAdapter.GenerateManifest(
		serviceDeployment,
		plan,
		requestParams,
		oldManifest,
		previousPlan,
		logger,
	)
}

func (m manifestGenerator) findPlans(planID string, previousPlanID *string) (serviceadapter.Plan, *serviceadapter.Plan, error) {
	plan, err := m.findPlan(planID)
	if err != nil {
		return serviceadapter.Plan{}, nil, err
	}

	if previousPlanID == nil {
		return plan, nil, nil
	}

	previousPlan, err := m.findPreviousPlan(*previousPlanID)
	if err != nil {
		return serviceadapter.Plan{}, nil, err
	}

	return plan, previousPlan, nil
}

func (m manifestGenerator) findPlan(planID string) (serviceadapter.Plan, error) {
	plan, found := m.serviceOffering.FindPlanByID(planID)
	if !found {
		return serviceadapter.Plan{}, NewDisplayableError(fmt.Errorf("plan %s does not exist", planID), fmt.Errorf("finding plan ID: %s", planID))
	}

	return serviceAdapterPlan(plan, m.serviceOffering.GlobalProperties), nil
}

func (m manifestGenerator) findPreviousPlan(previousPlanID string) (*serviceadapter.Plan, error) {
	previousPlan, found := m.serviceOffering.FindPlanByID(previousPlanID)
	if !found {
		return nil, NewDisplayableError(
			fmt.Errorf("previous plan %s does not exist", previousPlanID),
			fmt.Errorf("finding previous plan ID: %s", previousPlanID),
		)
	}

	convertedPlan := serviceAdapterPlan(previousPlan, m.serviceOffering.GlobalProperties)
	adapterPreviousPlan := &convertedPlan

	return adapterPreviousPlan, nil
}
