package broker_test

import (
	"fmt"
	"log"

	. "github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("ManifestGenerator", func() {
	Describe("GenerateManifest", func() {
		var (
			mg              ManifestGenerator
			serviceReleases serviceadapter.ServiceReleases
			serviceStemcell serviceadapter.Stemcell

			manifest []byte
			err      error

			deploymentName string
			instanceID     string
			requestParams  map[string]interface{}
			oldManifest    []byte
			planID         string
			previousPlanID *string
			logger         *log.Logger
		)

		BeforeEach(func() {
			//default valid params
			serviceReleases = serviceadapter.ServiceReleases{{
				Name:    "name",
				Version: "vers",
				Jobs:    []string{"a", "b"},
			}}
			serviceStemcell = serviceadapter.Stemcell{
				OS:      "RISC OS",
				Version: "2",
			}

			mg = NewManifestGenerator(
				serviceAdapter,
				serviceCatalog,
				serviceStemcell,
				serviceReleases,
			)
			deploymentName = "mydeployment"
			instanceID = "instance-id"
			requestParams = map[string]interface{}{"foo": "bar"}
			oldManifest = []byte("oldmanifest")
			planID = existingPlanID
			previousPlanID = nil
			logger = loggerFactory.NewLogger()
		})

		JustBeforeEach(func() {
			manifest, err = mg.GenerateManifest(
				deploymentName,
				instanceID,
				planID,
				requestParams,
				oldManifest,
				previousPlanID,
				logger,
			)
		})

		Context("when called with correct arguments", func() {
			generatedManifest := []byte("some manifest")
			BeforeEach(func() {
				serviceAdapter.GenerateManifestReturns(generatedManifest, nil)
			})

			It("calls service adapter once", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
			})

			It("returns result of adapter", func() {
				Expect(manifest).To(Equal(generatedManifest))
			})

			It("does not return an error", func() {
				Expect(err).To(Not(HaveOccurred()))
			})

			It("logs call to service adapter", func() {
				expectedLog := fmt.Sprintf("service adapter will generate manifest for instance %s\n", instanceID)
				Expect(logBuffer.String()).To(ContainSubstring(expectedLog))
			})

			It("calls the service adapter with the service deployment", func() {
				passedServiceDeployment, _, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				expectedServiceDeployment := serviceadapter.ServiceDeployment{
					DeploymentName: "mydeployment",
					Releases:       serviceReleases,
					Stemcell:       serviceStemcell,
				}
				Expect(passedServiceDeployment).To(Equal(expectedServiceDeployment))
			})

			It("calls the service adapter with the plan", func() {
				_, passedPlan, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(passedPlan.InstanceGroups).To(Equal(existingPlan.InstanceGroups))
			})

			It("calls the service adapter with the request params", func() {
				_, _, passedRequestParams, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(passedRequestParams).To(Equal(requestParams))
			})

			It("calls the service adapter with the old manifest", func() {
				_, _, _, passedOldManifest, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(passedOldManifest).To(Equal(oldManifest))
			})

			It("merges global and plan properties", func() {
				_, actualPlan, _, _, _, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				expectedProperties := serviceadapter.Properties{
					"a_global_property":          "global_value",
					"some_other_global_property": "other_global_value",
					"super": "no",
				}
				Expect(actualPlan.Properties).To(Equal(expectedProperties))
			})

			Context("when previous plan ID is provided", func() {
				BeforeEach(func() {
					anotherPlan := secondPlanID //cant & consts
					previousPlanID = &anotherPlan
				})

				It("calls the service adapter with the previous plan", func() {
					_, _, _, _, passedPreviousPlan, _ := serviceAdapter.GenerateManifestArgsForCall(0)
					Expect(passedPreviousPlan.InstanceGroups).To(Equal(secondPlan.InstanceGroups))
				})

				It("merges global and previous plan properties, overriding global with plan props", func() {
					_, _, _, _, previousPlan, _ := serviceAdapter.GenerateManifestArgsForCall(0)
					expectedProperties := serviceadapter.Properties{
						"a_global_property":          "overrides_global_value",
						"some_other_global_property": "other_global_value",
						"super": "yes",
					}

					Expect(previousPlan.Properties).To(Equal(expectedProperties))
				})
			})

			Context("when previous plan ID is not provided", func() {
				BeforeEach(func() {
					previousPlanID = nil
				})

				It("calls the service adapter with the nil previous plan", func() {
					_, _, _, _, passedPreviousPlan, _ := serviceAdapter.GenerateManifestArgsForCall(0)
					Expect(passedPreviousPlan).To(BeNil())
				})
			})
		})

		Context("when the plan cannot be found", func() {
			BeforeEach(func() {
				planID = "invalid-id"
			})

			It("does not call generate manifest", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(0))
			})

			It("returns an error", func() {
				Expect(err).To(MatchError(ContainSubstring("plan invalid-id does not exist")))
			})
		})

		Context("when the previous plan cannot be found", func() {
			BeforeEach(func() {
				invalidID := "invalid-previous-id"
				previousPlanID = &invalidID
			})

			It("does not call generate manifest", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(0))
			})

			It("returns an error", func() {
				Expect(err).To(MatchError(ContainSubstring("previous plan invalid-previous-id does not exist")))
			})
		})

		Context("when the adapter returns an error", func() {
			adapterError := fmt.Errorf("oops")
			BeforeEach(func() {
				serviceAdapter.GenerateManifestReturns(nil, adapterError)
			})

			It("is returned", func() {
				Expect(err).To(MatchError(adapterError))
			})
		})
	})

	Describe("ManifestsEqual", func() {
		manifestOne := BoshManifest(`---
age: old
name: floppy`)

		manifestTwo := BoshManifest(`---
age: old
name: smudge`)

		manifestThree := BoshManifest(`---
name: floppy
age: old`)

		invalidManifest := BoshManifest(`this is wrong`)

		Context("for two different manifests", func() {
			It("returns false", func() {
				result, _ := manifestOne.Equals(manifestTwo)
				Expect(result).To(BeFalse())
			})
		})

		Context("for the same manifests", func() {
			It("returns true", func() {
				result, _ := manifestOne.Equals(manifestOne)
				Expect(result).To(BeTrue())
			})
		})

		Context("for manifest with the same values in a different order", func() {
			It("returns true", func() {
				result, _ := manifestOne.Equals(manifestThree)
				Expect(result).To(BeTrue())
			})
		})

		Context("when first manifest is invalid yaml", func() {
			It("returns an error", func() {
				_, err := invalidManifest.Equals(manifestOne)
				Expect(err).To(HaveOccurred())
			})
		})

		Context("when second manifest is invalid yaml", func() {
			It("returns an error", func() {
				_, err := manifestOne.Equals(invalidManifest)
				Expect(err).To(HaveOccurred())
			})
		})
	})

})
