package broker

import (
	"context"
	"encoding/json"
	"fmt"
	"log"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

func (b *Broker) Provision(
	ctx context.Context,
	instanceID string,
	details brokerapi.ProvisionDetails,
	asyncAllowed bool,
) (brokerapi.ProvisionedServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeCreate), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	if !asyncAllowed {
		return brokerapi.ProvisionedServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	requestParams, err := convertProvisionDetailsToMap(details)
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	boshTaskID, boshContextID, dashboardURL, displayableErr := b.provisionInstance(
		ctx,
		instanceID,
		details,
		requestParams,
		logger,
	)
	if displayableErr.Occurred() {
		logger.Println(displayableErr)
		return brokerapi.ProvisionedServiceSpec{}, displayableErr.ErrorForCFUser()
	}

	operationData, err := json.Marshal(OperationData{
		BoshTaskID:    boshTaskID,
		OperationType: OperationTypeCreate,
		BoshContextID: boshContextID,
		PlanID:        details.PlanID,
	})
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	return brokerapi.ProvisionedServiceSpec{
		IsAsync:       true,
		DashboardURL:  dashboardURL,
		OperationData: string(operationData),
	}, nil
}

func (b *Broker) provisionInstance(
	ctx context.Context,
	instanceID string,
	details brokerapi.ProvisionDetails,
	requestParams map[string]interface{},
	logger *log.Logger,
) (int, string, string, DisplayableError) {
	errs := func(err DisplayableError) (int, string, string, DisplayableError) {
		return 0, "", "", err
	}

	plan, found := b.serviceOffering.FindPlanByID(details.PlanID)
	if !found {
		return errs(NewDisplayableError(fmt.Errorf("plan %s not found", details.PlanID), fmt.Errorf("finding plan ID %s", details.PlanID)))
	}

	manifestForDeployment, err := b.boshClient.GetManifestRaw(deploymentName(instanceID), logger)
	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("create", fmt.Errorf("could not get manifest: %s", err)))
	case error:
		return errs(NewGenericError(ctx, fmt.Errorf("could not get manifest: %s", err)))
	}

	if manifestForDeployment != nil {
		return errs(NewDisplayableError(brokerapi.ErrInstanceAlreadyExists, fmt.Errorf("deploying instance %s", instanceID)))
	}

	if err := b.validatePlanQuota(ctx, details.ServiceID, plan, logger); err != NilError {
		return errs(err)
	}

	var boshContextID string
	if plan.LifecycleErrands != nil {
		boshContextID = uuid.New()
	}

	boshTaskID, manifest, err := b.deployer.Deploy(instanceID, plan.ID, deploymentName(instanceID), OperationTypeCreate, requestParams, nil, nil, boshContextID, logger)
	switch err := err.(type) {
	case boshclient.RequestError:
		return errs(NewBoshRequestError("create", err))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	ctx = brokercontext.WithBoshTaskID(ctx, boshTaskID)

	dashboardUrl, err := b.serviceAdapter.GenerateDashboardUrl(instanceID, serviceAdapterPlan(plan, b.serviceOffering.GlobalProperties), manifest, logger)
	switch err := err.(type) {
	case ServiceAdapterCommandNotImplementedError:
		return boshTaskID, boshContextID, dashboardUrl, DisplayableError{}
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	return boshTaskID, boshContextID, dashboardUrl, DisplayableError{}
}
