package broker

import (
	"errors"
	"fmt"
	"log"
	"strings"

	"github.com/coreos/go-semver/semver"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

func (b *Broker) startupChecks() error {
	startupLogger := b.loggerFactory.NewLoggerWithoutRequestID()

	if err := b.checkAPIVersions(startupLogger); err != nil {
		return err
	}

	if err := b.verifyExistingInstancePlanIDsUnchanged(startupLogger); err != nil {
		return err
	}
	return nil
}

func (b *Broker) verifyExistingInstancePlanIDsUnchanged(logger *log.Logger) error {
	instanceCountByPlanID, err := b.cfClient.CountInstancesOfPlans(b.serviceOffering.ID, logger)
	if err != nil {
		return err
	}

	for planID, count := range instanceCountByPlanID {
		_, found := b.serviceOffering.Plans.FindByID(planID)

		if !found && count > 0 {
			return fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return nil
}

func (b *Broker) checkAPIVersions(startupLogger *log.Logger) error {
	var apiErrorMessages []string

	if err := checkCFAPIVersion(b.cfClient, startupLogger); err != nil {
		apiErrorMessages = append(apiErrorMessages, "CF API error: "+err.Error())
	}
	if err := checkBoshDirectorVersion(b.boshClient, startupLogger); err != nil {
		apiErrorMessages = append(apiErrorMessages, "BOSH Director error: "+err.Error())
	}

	if len(apiErrorMessages) > 0 {
		return errors.New(strings.Join(apiErrorMessages, " "))
	}

	return nil
}

func checkCFAPIVersion(cfClient CloudFoundryClient, logger *log.Logger) error {
	rawCFAPIVersion, err := cfClient.GetAPIVersion(logger)
	if err != nil {
		return err
	}

	version, err := semver.NewVersion(rawCFAPIVersion)
	if err != nil {
		return fmt.Errorf("Cloud Foundry API version couldn't be parsed. Expected a semver, got: %s.", rawCFAPIVersion)
	}
	if version.LessThan(*semver.New("2.57.0")) {
		return errors.New("Cloud Foundry API version is insufficient, ODB requires CF v238+.")
	}

	return nil
}

func checkBoshDirectorVersion(boshClient BoshClient, logger *log.Logger) error {
	rawDirectorVersion, err := boshClient.GetDirectorVersion(logger)

	switch err := err.(type) {
	case boshclient.UnrecognisedBoshDirectorVersionError:
		logger.Printf("Warning: %s. ODB requires BOSH v257+.", err)
		return nil
	case error:
		return err
	}

	if !rawDirectorVersion.SupportsODB() {
		return errors.New("BOSH Director API version is insufficient, ODB requires BOSH v257+.")
	}

	return nil
}
