package broker_test

import (
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

var _ = Describe("Initializing the broker", func() {
	Describe("check CF API version", func() {
		Context("when the CF API version is sufficient", func() {
			It("calls the CF API", func() {
				Expect(cfClient.GetAPIVersionCallCount()).To(Equal(1))
			})

			It("returns no error", func() {
				Expect(brokerCreationErr).NotTo(HaveOccurred())
			})
		})

		Context("when the CF API version is more than sufficient", func() {
			BeforeEach(func() {
				cfClient.GetAPIVersionReturns("3.0.0", nil)
			})

			It("calls the CF API", func() {
				Expect(cfClient.GetAPIVersionCallCount()).To(Equal(1))
			})

			It("returns no error", func() {
				Expect(brokerCreationErr).NotTo(HaveOccurred())
			})
		})

		Context("when the CF API version is insufficient", func() {
			BeforeEach(func() {
				cfClient.GetAPIVersionReturns("2.56.0", nil)
			})

			It("calls the CF API", func() {
				Expect(cfClient.GetAPIVersionCallCount()).To(Equal(1))
			})

			It("returns no error", func() {
				Expect(brokerCreationErr).To(HaveOccurred())
				Expect(brokerCreationErr.Error()).To(ContainSubstring("Cloud Foundry API version is insufficient, ODB requires CF v238+"))
			})
		})

		Context("when getting the CF API version fails", func() {
			BeforeEach(func() {
				cfClient.GetAPIVersionReturns("", fmt.Errorf("nothing today, thank you"))
			})

			It("calls the CF API", func() {
				Expect(cfClient.GetAPIVersionCallCount()).To(Equal(1))
			})

			It("returns an error", func() {
				Expect(brokerCreationErr).To(HaveOccurred())
				Expect(brokerCreationErr.Error()).To(ContainSubstring("nothing today, thank you"))
			})
		})

		Context("when the CF API version cannot be parsed", func() {
			BeforeEach(func() {
				cfClient.GetAPIVersionReturns("1.wrong.0", nil)
			})

			It("calls the CF API", func() {
				Expect(cfClient.GetAPIVersionCallCount()).To(Equal(1))
			})

			It("returns an error", func() {
				Expect(brokerCreationErr).To(HaveOccurred())
				Expect(brokerCreationErr.Error()).To(ContainSubstring("Cloud Foundry API version couldn't be parsed. Expected a semver, got: 1.wrong.0"))
			})
		})
	})

	Describe("check BOSH director version", func() {
		Context("when the BOSH director version supports ODB", func() {
			BeforeEach(func() {
				boshDirectorVersion = boshclient.NewBoshDirectorVersion(boshclient.MinimumMajorStemcellDirectorVersionForODB, boshclient.StemcellDirectorVersionType)
				boshClient.GetDirectorVersionReturns(boshDirectorVersion, nil)
			})

			It("calls the BOSH director", func() {
				Expect(boshClient.GetDirectorVersionCallCount()).To(Equal(1))
			})

			It("does not return an error", func() {
				Expect(brokerCreationErr).NotTo(HaveOccurred())
			})
		})

		Context("when the BOSH director version does not support ODB", func() {
			BeforeEach(func() {
				boshDirectorVersion = boshclient.NewBoshDirectorVersion(boshclient.MinimumMajorStemcellDirectorVersionForODB-1, boshclient.StemcellDirectorVersionType)
				boshClient.GetDirectorVersionReturns(boshDirectorVersion, nil)
			})

			It("calls the BOSH director", func() {
				Expect(boshClient.GetDirectorVersionCallCount()).To(Equal(1))
			})

			It("returns an error", func() {
				Expect(brokerCreationErr).To(HaveOccurred())
				Expect(brokerCreationErr.Error()).To(ContainSubstring("BOSH Director API version is insufficient, ODB requires BOSH v257+"))
			})
		})

		Context("when getting the BOSH director version fails", func() {
			BeforeEach(func() {
				boshClient.GetDirectorVersionReturns(boshclient.BoshDirectorVersion{}, fmt.Errorf("nothing today, thank you"))
			})

			It("calls the BOSH director", func() {
				Expect(boshClient.GetDirectorVersionCallCount()).To(Equal(1))
			})

			It("returns an error", func() {
				Expect(brokerCreationErr).To(HaveOccurred())
				Expect(brokerCreationErr.Error()).To(ContainSubstring("nothing today, thank you"))
			})
		})

		Context("when the BOSH director version is unrecognised (e.g. bosh director 260.3)", func() {
			BeforeEach(func() {
				boshClient.GetDirectorVersionReturns(boshclient.BoshDirectorVersion{}, boshclient.NewUnrecognisedBoshDirectorVersionError("0000 (00000000)"))
			})

			It("calls the BOSH director", func() {
				Expect(boshClient.GetDirectorVersionCallCount()).To(Equal(1))
			})

			It("does not return an error", func() {
				Expect(brokerCreationErr).NotTo(HaveOccurred())
			})

			It("logs a warning", func() {
				Expect(logBuffer.String()).To(ContainSubstring(`Warning: unrecognised BOSH Director version: "0000 (00000000)". ODB requires BOSH v257+.`))
			})
		})
	})

	Describe("check CF service instances", func() {
		Context("when there are no pre-existing instances", func() {
			BeforeEach(func() {
				cfClient.GetAPIVersionReturns("2.57.0", nil)
				cfClient.CountInstancesOfPlansReturns(map[string]int{}, nil)
			})

			It("returns no error", func() {
				Expect(brokerCreationErr).NotTo(HaveOccurred())
			})
		})

		Context("when there are no pre-existing instances of configured plans", func() {
			Context("service catalog contains the existing instance's plan", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlansReturns(map[string]int{existingPlanID: 1}, nil)
				})

				It("returns no error", func() {
					Expect(brokerCreationErr).NotTo(HaveOccurred())
				})
			})

			Context("service catalog does not contain a plan with zero instances", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlansReturns(map[string]int{existingPlanID: 1, "non-existing-plan": 0}, nil)
				})

				It("returns no error", func() {
					Expect(brokerCreationErr).NotTo(HaveOccurred())
				})
			})

			Context("service catalog does not contain the existing instance's plan", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlansReturns(map[string]int{"i-dont-exist-in-the-catalog": 1}, nil)
				})

				It("returns an error", func() {
					Expect(brokerCreationErr).To(MatchError(ContainSubstring("You cannot change the plan_id of a plan that has existing service instances")))
				})
			})

			Context("when instances cannot be retrieved", func() {
				BeforeEach(func() {
					cfClient.CountInstancesOfPlansReturns(nil, errors.New("flurp"))
				})

				It("returns an error", func() {
					Expect(brokerCreationErr).To(HaveOccurred())
				})
			})
		})
	})

})
