package broker

import (
	"context"
	"errors"
	"fmt"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
	"github.com/pivotal-cf/on-demand-services-sdk/serviceadapter"
)

func (b *Broker) Unbind(
	ctx context.Context,
	instanceID,
	bindingID string,
	details brokerapi.UnbindDetails,
) error {
	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeUnbind), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	errs := func(err DisplayableError) error {
		logger.Println(err)
		return err.ErrorForCFUser()
	}

	vms, manifest, err := b.getDeploymentInfo(instanceID, logger)
	switch err.(type) {
	case boshclient.DeploymentNotFoundError:
		return errs(NewDisplayableError(brokerapi.ErrInstanceDoesNotExist, fmt.Errorf("error unbinding: instance %s, not found", instanceID)))
	case error:
		return errs(NewGenericError(ctx, fmt.Errorf("gathering unbinding info %s", err)))
	}

	requestParams := map[string]interface{}{
		"plan_id":    details.PlanID,
		"service_id": details.ServiceID,
	}

	logger.Printf("service adapter will delete binding with ID %s for instance %s\n", bindingID, instanceID)
	err = b.serviceAdapter.DeleteBinding(bindingID, vms, manifest, requestParams, logger)
	switch err := err.(type) {
	case serviceadapter.BindingNotFoundError:
		return errs(NewDisplayableError(brokerapi.ErrBindingDoesNotExist, errors.New("deleting binding")))
	case DisplayableError:
		return errs(err)
	case error:
		return errs(NewGenericError(ctx, err))
	}

	return nil
}
