package broker

import (
	"fmt"
	"log"

	"github.com/pborman/uuid"
)

func (b *Broker) Upgrade(instanceID string, logger *log.Logger) (OperationData, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	instance, err := b.cfClient.GetInstanceState(instanceID, logger)
	if err != nil {
		if IsNotFound(err) {
			return OperationData{}, NewDeploymentNotFoundError(err)
		}
		return OperationData{}, err
	}

	// Ensure last operation is "complete". This will already take lifecycle errands into account.
	if instance.OperationInProgress {
		return OperationData{}, NewOperationInProgressError(fmt.Errorf("cloud controller: operation in progress for instance %s", instanceID))
	}

	logger.Printf("upgrading instance %s", instanceID)

	plan, found := b.serviceOffering.FindPlanByID(instance.PlanID)
	if !found {
		logger.Printf("error: finding plan ID %s", instance.PlanID)
		return OperationData{}, fmt.Errorf("plan %s not found", instance.PlanID)
	}

	var boshContextID string
	if plan.LifecycleErrands != nil {
		boshContextID = uuid.New()
	}

	taskID, err := b.redeploy(instanceID, instance.PlanID, OperationTypeUpgrade, nil, &instance.PlanID, boshContextID, logger)
	if err != nil {
		logger.Printf("error upgrading instance %s: %s", instanceID, err)

		displayable, ok := err.(DisplayableError)
		if ok {
			return OperationData{}, displayable.errorForOperator
		}

		return OperationData{}, err
	}

	return OperationData{
		BoshContextID: boshContextID,
		BoshTaskID:    taskID,
		PlanID:        instance.PlanID,
		OperationType: OperationTypeUpgrade,
	}, nil
}
