package broker_test

import (
	"errors"
	"fmt"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("Upgrade", func() {
	var (
		upgradeOperationData broker.OperationData
		instanceID           = "some-instance"
		logger               *log.Logger

		context = ItRedeploysContext{
			instanceID:    instanceID,
			oldPlanID:     existingPlanID,
			newPlanID:     existingPlanID,
			operationType: broker.OperationTypeUpgrade,
			boshTaskID:    876,
		}
	)

	JustBeforeEach(func() {
		logger = loggerFactory.NewLogger()
		upgradeOperationData, context.redeployErr = b.Upgrade(context.instanceID, logger)
	})

	ItRedeploys(&context)

	Context("when there is a previous deployment for the service instance", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			fakeDeployer.DeployReturns(context.boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns the task ID for the upgrade task", func() {
			Expect(upgradeOperationData.BoshTaskID).To(Equal(context.boshTaskID))
		})

		It("returns an empty context id", func() {
			Expect(upgradeOperationData.BoshContextID).To(BeEmpty())
		})

		It("returns the plan id", func() {
			Expect(upgradeOperationData.PlanID).To(Equal(existingPlanID))
		})

		It("returns the operation type", func() {
			Expect(upgradeOperationData.OperationType).To(Equal(broker.OperationTypeUpgrade))
		})

		It("fetches correct instance", func() {
			Expect(cfClient.GetInstanceStateCallCount()).To(Equal(1))
			actualInstanceID, _ := cfClient.GetInstanceStateArgsForCall(0)
			Expect(actualInstanceID).To(Equal(instanceID))
		})

		It("logs with a request ID", func() {
			Expect(logBuffer.String()).To(MatchRegexp(`\[[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\] \d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} upgrading instance`))
		})

		Context("and there is an operation in progress on the instance", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID, OperationInProgress: true}, nil)
			})

			It("returns an OperationInProgressError", func() {
				Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
			})
		})

		Context("and post-deploy errand is configured", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: postDeployErrandPlanID}, nil)
			})

			It("deploys with a context id", func() {
				_, _, _, _, _, _, _, contextID, _ := fakeDeployer.DeployArgsForCall(0)
				Expect(contextID).ToNot(BeEmpty())
			})

			It("returns the correct context id", func() {
				_, _, _, _, _, _, _, contextID, _ := fakeDeployer.DeployArgsForCall(0)
				Expect(upgradeOperationData.BoshContextID).To(Equal(contextID))
			})
		})
	})

	Context("when the service instance does not exist in Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, broker.NewNotFoundError())
		})

		It("returns a DeploymentNotFoundError", func() {
			Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.DeploymentNotFoundError{}))
		})
	})

	Context("when the instance state cannot be retrieved from Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, errors.New("no instance for you"))
		})

		It("returns the error to the cf user", func() {
			Expect(context.redeployErr).To(MatchError(ContainSubstring("no instance for you")))
		})
	})

	Context("when the plan cannot be found", func() {
		BeforeEach(func() {
			context.newPlanID = "non-existant-plan-id"
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: context.newPlanID}, nil)
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)

		})
		It("returns a user displayable error", func() {
			Expect(context.redeployErr).To(MatchError(ContainSubstring(fmt.Sprintf("plan %s not found", context.newPlanID))))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error: finding plan ID %s", context.newPlanID)))
		})

		It("does not redeploy", func() {
			Expect(serviceAdapter.GenerateManifestCallCount()).To(BeZero())
			Expect(boshClient.GetManifestRawCallCount()).To(BeZero())
			Expect(fakeDeployer.DeployCallCount()).To(BeZero())
		})
	})

	Context("when there is a task in progress on the instance", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			fakeDeployer.DeployReturns(0, []byte{}, broker.NewOperationInProgressError(fmt.Errorf("deployment %s is still in progress", deploymentName(instanceID))))
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns an OperationInProgressError", func() {
			Expect(context.redeployErr).To(BeAssignableToTypeOf(broker.OperationInProgressError{}))
		})
	})
})
