package upgrader

import (
	"encoding/json"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

type upgrader struct {
	brokerUsername  string
	brokerPassword  string
	brokerUrl       string
	pollingInterval time.Duration
}

func New(brokerUsername, brokerPassword, brokerUrl string, pollingIntervalMilliseconds int64) upgrader {
	return upgrader{
		brokerUsername:  brokerUsername,
		brokerPassword:  brokerPassword,
		brokerUrl:       brokerUrl,
		pollingInterval: time.Duration(pollingIntervalMilliseconds) * time.Millisecond,
	}
}

func (u upgrader) Upgrade() (upgraded []string, skipped []string, err error) {
	errs := func(err error) ([]string, []string, error) {
		return nil, nil, err
	}

	instances, err := u.getServiceInstances()
	if err != nil {
		return errs(fmt.Errorf("error getting service instances: %s\n", err))
	}

	for len(instances) != 0 {
		upgradedInstances, skippedInstances, retryInstances, err := u.upgradeInstances(instances)
		if err != nil {
			return errs(err)
		}

		upgraded = append(upgraded, upgradedInstances...)
		skipped = append(skipped, skippedInstances...)
		instances = retryInstances

		if len(instances) > 0 {
			time.Sleep(u.pollingInterval)
		}
	}

	return
}

func (u upgrader) upgradeInstances(instances []string) (upgradedInstances []string, skippedInstances []string, retryInstances []string, err error) {
	errs := func(err error) ([]string, []string, []string, error) {
		return nil, nil, nil, err
	}

	for _, instance := range instances {
		resp, skip, retry, err := u.upgradeInstance(instance)
		if err != nil {
			return errs(fmt.Errorf("Upgrade failed for service_instance_id %s: %s\n", instance, err))
		}

		if skip {
			skippedInstances = append(skippedInstances, instance)
			continue
		}
		if retry {
			retryInstances = append(retryInstances, instance)
			continue
		}

		if err := u.pollLastOperation(instance, resp); err != nil {
			return errs(err)
		}

		upgradedInstances = append(upgradedInstances, instance)
	}
	return
}

func (u upgrader) pollLastOperation(instance string, opData broker.OperationData) error {
	for {
		time.Sleep(u.pollingInterval)

		lastOperation, err := u.getLastOperation(instance, opData)
		if err != nil {
			return fmt.Errorf("error getting last operation: %s\n", err)
		}

		switch lastOperation.State {
		case brokerapi.Failed:
			return fmt.Errorf("Upgrade failed for service_instance_id %s: %s", instance, lastOperation.Description)
		case brokerapi.Succeeded:
			return nil
		}
	}
}

func (u upgrader) getServiceInstances() ([]string, error) {
	var instances []mgmtapi.Instance
	if err := u.makeRequest("/mgmt/service_instances", &instances); err != nil {
		return nil, err
	}

	var instanceIDs []string
	for _, instance := range instances {
		instanceIDs = append(instanceIDs, instance.InstanceID)
	}
	return instanceIDs, nil
}

func (u upgrader) upgradeInstance(instance string) (opData broker.OperationData, skip bool, retry bool, err error) {
	request, err := http.NewRequest("PATCH", fmt.Sprintf("%s%s", u.brokerUrl, "/mgmt/service_instances/"+instance), nil)
	if err != nil {
		skip, retry = false, false
		return
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		skip, retry = false, false
		return
	}
	defer response.Body.Close()

	switch response.StatusCode {
	case http.StatusAccepted:
		skip, retry = false, false

		if err = json.NewDecoder(response.Body).Decode(&opData); err != nil {
			err = fmt.Errorf("can't parse response %s", err)
		}

		return
	case http.StatusGone:
		skip, retry, err = true, false, nil
		return
	case http.StatusConflict:
		skip, retry, err = false, true, nil
		return
	default:
		skip, retry = false, false

		var errorResponse brokerapi.ErrorResponse
		if err = json.NewDecoder(response.Body).Decode(&errorResponse); err != nil {
			err = fmt.Errorf("can't parse response %s", err)
			return
		}

		err = fmt.Errorf("ERROR: %s", errorResponse.Description)
		return
	}
}

func (u upgrader) getLastOperation(instance string, opData broker.OperationData) (brokerapi.LastOperation, error) {
	opDataJSON, err := json.Marshal(opData)
	if err != nil {
		return brokerapi.LastOperation{}, err
	}

	operationQueryStr := url.QueryEscape(string(opDataJSON))

	var lastOperation brokerapi.LastOperation
	if err := u.makeRequest("/v2/service_instances/"+instance+"/last_operation?operation="+operationQueryStr, &lastOperation); err != nil {
		return brokerapi.LastOperation{}, err
	}

	return lastOperation, nil
}

func (u upgrader) makeRequest(requestPath string, responseData interface{}) error {
	request, err := http.NewRequest("GET", fmt.Sprintf("%s%s", u.brokerUrl, requestPath), nil)
	if err != nil {
		return err
	}

	request.SetBasicAuth(u.brokerUsername, u.brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return err
	}

	if response.StatusCode != 200 {
		return fmt.Errorf("response returned bad status: %s", response.Status)
	}

	err = json.NewDecoder(response.Body).Decode(responseData)
	if err != nil {
		return err
	}

	return nil
}
